from typing import Any

from pyModeS import adsb, bds, commb, py_common  # type: ignore


def decode(msg: str, common: Any = py_common) -> dict[str, Any]:
    """rs1090.decode coded in pyModeS

    This is a rough equivalement of the rs1090 decode function (with few
    fields left undecoded). Let's say it's not a big deal.

    There **must** be some imprecisions, but that's enough for benchmarking.

    """

    df = common.df(msg)
    icao = common.icao(msg)

    decoded = {"msg": msg, "icao": icao, "df": df}

    if df == 17:
        decoded["tc"] = tc = common.typecode(msg)

        if tc is None:
            return decoded

        if 1 <= tc <= 4:  # callsign
            callsign = adsb.callsign(msg)
            decoded["bds"] = 10
            decoded["callsign"] = callsign

        if 5 <= tc <= 8:  # surface position
            decoded["bds"] = "06"
            oe = adsb.oe_flag(msg)
            msgbin = common.hex2bin(msg)
            cprlat = common.bin2int(msgbin[54:71]) / 131072.0
            cprlon = common.bin2int(msgbin[71:88]) / 131072.0
            v = adsb.surface_velocity(msg)
            decoded["oddflag"] = "odd" if oe else "even"
            decoded["cprlat"] = cprlat
            decoded["cprlon"] = cprlon
            decoded["speed"] = v[0]
            decoded["track"] = v[1]

        if 9 <= tc <= 18:  # airborne position
            decoded["bds"] = "05"
            alt = adsb.altitude(msg)
            oe = adsb.oe_flag(msg)
            msgbin = common.hex2bin(msg)
            cprlat = common.bin2int(msgbin[54:71]) / 131072.0
            cprlon = common.bin2int(msgbin[71:88]) / 131072.0
            decoded["oddflag"] = "odd" if oe else "even"
            decoded["cprlat"] = cprlat
            decoded["cprlon"] = cprlon
            decoded["altitude"] = alt

        if tc == 19:
            decoded["bds"] = "09"
            velocity = adsb.velocity(msg)
            if velocity is not None:
                spd, trk, vr, t = velocity
                types = {"GS": "Ground speed", "TAS": "True airspeed"}
                decoded[types[t]] = spd
                decoded["track"] = trk
                decoded["vertical_rate"] = vr

        if 20 <= tc <= 22:  # airborne position
            decoded["bds"] = "05"
            alt = adsb.altitude(msg)
            oe = adsb.oe_flag(msg)
            msgbin = common.hex2bin(msg)
            cprlat = common.bin2int(msgbin[54:71]) / 131072.0
            cprlon = common.bin2int(msgbin[71:88]) / 131072.0
            decoded["oddflag"] = "odd" if oe else "even"
            decoded["cprlat"] = cprlat
            decoded["cprlon"] = cprlon
            decoded["altitude"] = alt

        if tc == 29:  # target state and status
            decoded["bds"] = "62"
            subtype = common.bin2int((common.hex2bin(msg)[32:])[5:7])
            tcas_operational = adsb.tcas_operational(msg)
            types_29 = {0: "Not Engaged", 1: "Engaged"}
            tcas_operational_types = {0: "Not Operational", 1: "Operational"}
            if subtype == 0:
                emergency_types = {
                    0: "No emergency",
                    1: "General emergency",
                    2: "Lifeguard/medical emergency",
                    3: "Minimum fuel",
                    4: "No communications",
                    5: "Unlawful interference",
                    6: "Downed aircraft",
                    7: "Reserved",
                }
                vertical_horizontal_types = {
                    1: "Acquiring mode",
                    2: "Capturing/Maintaining mode",
                }
                tcas_ra_types = {0: "Not active", 1: "Active"}
                alt, alt_source, alt_ref = adsb.target_altitude(msg)
                angle, angle_type, angle_source = adsb.target_angle(msg)
                vertical_mode = adsb.vertical_mode(msg)
                horizontal_mode = adsb.horizontal_mode(msg)
                tcas_ra = adsb.tcas_ra(msg)
                emergency_status = adsb.emergency_status(msg)
                decoded["target_alt"] = alt
                decoded["alt_source"] = alt_source
                decoded["alt_reference"] = alt_ref
                decoded["angle"] = angle
                decoded["angle_type"] = angle_type
                decoded["angle_source"] = angle_source
                if vertical_mode is not None:
                    decoded["vnav_mode"] = (
                        vertical_horizontal_types[vertical_mode],
                    )

                if horizontal_mode is not None:
                    decoded["lnav_mode"] = (
                        vertical_horizontal_types[horizontal_mode],
                    )
                decoded["tcas"] = (
                    tcas_operational_types[tcas_operational]
                    if tcas_operational
                    else None,
                )
                decoded["tcas_type"] = tcas_ra_types[tcas_ra]
                decoded["emergency_status"] = emergency_types[emergency_status]
            else:
                alt, alt_source = adsb.selected_altitude(msg)
                baro = adsb.baro_pressure_setting(msg)
                hdg = adsb.selected_heading(msg)
                autopilot = adsb.autopilot(msg)
                vnav = adsb.vnav_mode(msg)
                alt_hold = adsb.altitude_hold_mode(msg)
                app = adsb.approach_mode(msg)
                lnav = adsb.lnav_mode(msg)
                decoded["selected_alt"] = alt
                decoded["alt_source"] = alt_source
                decoded["barometric_setting"] = baro
                decoded["selected_hdg"] = hdg
                if not (common.bin2int((common.hex2bin(msg)[32:])[46]) == 0):
                    decoded["autopilot"] = (
                        types_29[autopilot] if autopilot else None
                    )
                    decoded["vnav_mode"] = types_29[vnav] if vnav else None
                    decoded["alt_hold"] = (
                        types_29[alt_hold] if alt_hold else None,
                    )
                    decoded["app_mode"] = types_29[app] if app else None
                    decoded["tcas"] = (
                        tcas_operational_types[tcas_operational]
                        if tcas_operational
                        else None,
                    )
                    decoded["lnav_mode"] = types_29[lnav] if lnav else None

    if df == 20:
        decoded["altitude"] = common.altcode(msg)

    if df == 21:
        decoded["squawk"] = common.idcode(msg)

    if df == 20 or df == 21:
        decoded["bds"] = BDS = bds.infer(msg, mrar=True)

        if BDS == "BDS20":
            decoded["callsign"] = callsign = commb.cs20(msg)

        if BDS == "BDS40":
            decoded["selected_mcp"] = commb.selalt40mcp(msg)
            decoded["selected_fms"] = commb.selalt40fms(msg)
            decoded["selected_baro"] = commb.p40baro(msg)

        if BDS == "BDS50":
            decoded["Roll angle"] = commb.roll50(msg)
            decoded["Track angle"] = commb.trk50(msg)
            decoded["Track rate"] = commb.rtrk50(msg)
            decoded["Ground speed"] = commb.gs50(msg)
            decoded["True airspeed"] = commb.tas50(msg)

        if BDS == "BDS60":
            decoded["Megnatic Heading"] = commb.hdg60(msg)
            decoded["Indicated airspeed"] = commb.ias60(msg)
            decoded["Mach number"] = commb.mach60(msg)
            decoded["Vertical rate (Baro)"] = commb.vr60baro(msg)
            decoded["Vertical rate (INS)"] = commb.vr60ins(msg)

        if BDS == "BDS44":
            decoded["Wind speed"] = commb.wind44(msg)[0]
            decoded["Wind direction"] = commb.wind44(msg)[1]
            decoded["Temperature 1"] = commb.temp44(msg)[0]
            decoded["Temperature 2"] = commb.temp44(msg)[1]
            decoded["Pressure"] = commb.p44(msg)
            decoded["Humidity"] = commb.hum44(msg)
            decoded["Turbulence"] = commb.turb44(msg)

        if BDS == "BDS45":
            decoded["Turbulence"] = commb.turb45(msg)
            decoded["Wind shear"] = commb.ws45(msg)
            decoded["Microbust"] = commb.mb45(msg)
            decoded["Icing"] = commb.ic45(msg)
            decoded["Wake vortex"] = commb.wv45(msg)
            decoded["Temperature"] = commb.temp45(msg)
            decoded["Pressure"] = commb.p45(msg)
            decoded["Radio height"] = commb.rh45(msg)

    return decoded



from typing import Sequence

from .stubs import DF17_BDS05, DF17_BDS65, DF20, DF21

def aircraft_information(
    icao24: str, registration: None | str = None
) -> dict[str, str]: ...
def decode_1090(msg: str) -> list[int]: ...
def decode_1090_with_reference(
    msg: str, reference: tuple[float, float]
) -> list[int]: ...
def decode_1090_vec(msgs: Sequence[Sequence[str]]) -> list[int]: ...
def decode_1090t_vec(
    msgs: Sequence[Sequence[str]],
    ts: Sequence[Sequence[float]],
    reference: None | tuple[float, float] = None,
) -> list[int]: ...
def decode_flarm(
    msg: str, timestamp: int, reflat: float, reflon: float
) -> list[int]: ...
def decode_flarm_vec(
    msgs: Sequence[Sequence[str]],
    ts: Sequence[Sequence[int]],
    reflat: Sequence[Sequence[float]],
    reflon: Sequence[Sequence[float]],
) -> list[int]: ...
def decode_bds05(mgs: str) -> DF17_BDS05: ...
def decode_bds10(mgs: str) -> DF20 | DF21: ...
def decode_bds17(mgs: str) -> DF20 | DF21: ...
def decode_bds18(mgs: str) -> DF20 | DF21: ...
def decode_bds19(mgs: str) -> DF20 | DF21: ...
def decode_bds20(mgs: str) -> DF20 | DF21: ...
def decode_bds21(mgs: str) -> DF20 | DF21: ...
def decode_bds30(mgs: str) -> DF20 | DF21: ...
def decode_bds40(mgs: str) -> DF20 | DF21: ...
def decode_bds44(mgs: str) -> DF20 | DF21: ...
def decode_bds45(mgs: str) -> DF20 | DF21: ...
def decode_bds50(mgs: str) -> DF20 | DF21: ...
def decode_bds60(mgs: str) -> DF20 | DF21: ...
def decode_bds65(mgs: str) -> DF17_BDS65: ...



# ruff: noqa: E402
# %%
import pandas as pd  # type: ignore

from rs1090 import decode

data = pd.read_csv(
    "../../crates/rs1090/data/long_flight.csv",
    names=["timestamp", "rawmsg"],
)

# %%
decoded = decode(data.rawmsg.str[18:], data.timestamp, reference=(43.3, 1.35))

# %%
df = pd.DataFrame.from_records(decoded)
df = df.assign(timestamp=pd.to_datetime(df.timestamp, unit="s", utc=True))
df

# %%
from itertools import batched

# 5000 is quite fastest actually!
df = pd.concat(pd.DataFrame.from_records(d) for d in batched(decoded, 5000))
df = df.assign(timestamp=pd.to_datetime(df.timestamp, unit="s", utc=True))
df

# %%
from traffic.core import Traffic  # type: ignore

t = Traffic(df)
t
# %%
t["3944ed"].map_leaflet()
# %%
t["486257"].map_leaflet()
# %%



# ruff: noqa: E402
# %%
import pandas as pd  # type: ignore

from rs1090 import flarm

data = pd.read_csv("../../crates/rs1090/data/flarm.csv")

# %%

decoded = flarm(
    data.rawmessage,
    data.timestamp.astype(int),
    data.sensorlatitude,
    data.sensorlongitude,
)

# %%
from itertools import batched

# 5000 is a good batch size for fast loading!
df = pd.concat(pd.DataFrame.from_records(d) for d in batched(decoded, 5000))
df = df.assign(timestamp=pd.to_datetime(df.timestamp, unit="s", utc=True))
df

# %%
# More advanced/expressive preprocessing available in the traffic library
# https://github.com/xoolive/traffic

from pitot.geodesy import distance  # type: ignore

flight = df.query(
    'icao24 == "38f27b" and '
    '"2022-06-15 07:35Z" < timestamp < "2022-06-15 09:30Z"'
).sort_values("timestamp")

flight = flight.assign(
    distance=distance(
        flight.latitude,
        flight.longitude,
        flight.reference_lat,
        flight.reference_lon,
    ),
)
# Remove noisy points (impossible distance to the receiver)
flight = flight.query("distance < 200_000")

# %%
import numpy as np

coords = flight[["timestamp", "latitude", "longitude"]]
delta = pd.concat([coords, coords.add_suffix("_1").diff()], axis=1)
delta_1 = delta.iloc[1:]
distance_nm = distance(
    (delta_1.latitude - delta_1.latitude_1).values,
    (delta_1.longitude - delta_1.longitude_1).values,
    delta_1.latitude.values,
    delta_1.longitude.values,
)
secs = delta_1.timestamp_1.dt.total_seconds()

# Remove irrealistic jumps from a point to another
flight = flight.assign(
    gs=list(np.abs(np.pad(distance_nm / secs, (1, 0), "edge")))
)


# %%
import matplotlib.pyplot as plt  # type: ignore
from cartes.crs import Lambert93, PlateCarree  # type: ignore
from cartes.osm import Overpass  # type: ignore

# Get the airport layout
airport = Overpass.request(area=dict(icao="LFMY"), aeroway=True)

fig, ax = plt.subplots(subplot_kw=dict(projection=Lambert93()))
airport.plot(
    ax,
    by="aeroway",
    runway={"lw": 2},
    aerodrome={"alpha": 0},
    radar={"alpha": 0},
)
flight.query("gs < 20000").plot(
    ax=ax, x="longitude", y="latitude", legend=False, transform=PlateCarree()
)
ax.spines["geo"].set_visible(False)
ax.yaxis.set_visible(False)
ax.set_extent((5.1, 5.16, 43.59, 43.64))  # type: ignore
fig.savefig("flarm.png")

# %%



from __future__ import annotations

import pickle
from typing import (
    Any,
    Callable,
    Iterable,
    ParamSpec,
    Sequence,
    TypeVar,
    overload,
)

import pandas as pd

from ._rust import (
    aircraft_information,
    decode_1090,
    decode_1090_vec,
    decode_1090_with_reference,
    decode_1090t_vec,
    decode_bds05,
    decode_bds10,
    decode_bds17,
    decode_bds18,
    decode_bds19,
    decode_bds20,
    decode_bds21,
    decode_bds30,
    decode_bds40,
    decode_bds44,
    decode_bds45,
    decode_bds50,
    decode_bds60,
    decode_bds65,
    decode_flarm,
    decode_flarm_vec,
)
from .stubs import (
    Flarm,
    Message,
    is_bds05,
    is_bds06,
    is_bds08,
    is_bds09,
    is_bds10,
    is_bds17,
    is_bds20,
    is_bds30,
    is_bds40,
    is_bds44,
    is_bds50,
    is_bds60,
    is_bds61,
    is_bds62,
    is_bds65,
    is_df0,
    is_df4,
    is_df5,
    is_df11,
    is_df16,
    is_df17,
    is_df18,
    is_df20,
    is_df21,
)

try:
    # new in Python 3.12
    from itertools import batched  # type: ignore[attr-defined]
except ImportError:
    from itertools import islice

    T = TypeVar("T")

    def batched(iterable: Sequence[T], n: int) -> Iterable[tuple[T, ...]]:
        # batched('ABCDEFG', 3) --> ABC DEF G
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(islice(it, n)):
            yield batch


P = ParamSpec("P")


def unpickle_fun(fun: Callable[P, list[int]]) -> Callable[P, Any]:
    """Wrapper that unpickles the return value from Rust functions.

    The Rust bindings return pickled objects as list[int], which this
    wrapper converts back to the actual Python types.
    """

    def wrapped_fun(*args: P.args, **kwargs: P.kwargs) -> Any:
        int_list = fun(*args, **kwargs)
        return pickle.loads(bytes(int_list))

    return wrapped_fun


decode_bds05 = unpickle_fun(decode_bds05)  # type: ignore[arg-type]
decode_bds10 = unpickle_fun(decode_bds10)  # type: ignore[arg-type]
decode_bds17 = unpickle_fun(decode_bds17)  # type: ignore[arg-type]
decode_bds18 = unpickle_fun(decode_bds18)  # type: ignore[arg-type]
decode_bds19 = unpickle_fun(decode_bds19)  # type: ignore[arg-type]
decode_bds20 = unpickle_fun(decode_bds20)  # type: ignore[arg-type]
decode_bds21 = unpickle_fun(decode_bds21)  # type: ignore[arg-type]
decode_bds30 = unpickle_fun(decode_bds30)  # type: ignore[arg-type]
decode_bds40 = unpickle_fun(decode_bds40)  # type: ignore[arg-type]
decode_bds44 = unpickle_fun(decode_bds44)  # type: ignore[arg-type]
decode_bds45 = unpickle_fun(decode_bds45)  # type: ignore[arg-type]
decode_bds50 = unpickle_fun(decode_bds50)  # type: ignore[arg-type]
decode_bds60 = unpickle_fun(decode_bds60)  # type: ignore[arg-type]
decode_bds65 = unpickle_fun(decode_bds65)  # type: ignore[arg-type]


__all__ = [
    "Flarm",
    "Message",
    "batched",
    "decode",
    "decode_bds05",
    "decode_bds10",
    "decode_bds17",
    "decode_bds20",
    "decode_bds21",
    "decode_bds30",
    "decode_bds40",
    "decode_bds44",
    "decode_bds45",
    "decode_bds50",
    "decode_bds60",
    "flarm",
    "is_bds05",
    "is_bds06",
    "is_bds08",
    "is_bds09",
    "is_bds10",
    "is_bds17",
    "is_bds20",
    "is_bds30",
    "is_bds40",
    "is_bds44",
    "is_bds50",
    "is_bds60",
    "is_bds61",
    "is_bds62",
    "is_bds65",
    "is_df0",
    "is_df11",
    "is_df16",
    "is_df17",
    "is_df18",
    "is_df20",
    "is_df21",
    "is_df4",
    "is_df5",
    "aircraft_information",
]


@overload
def decode(
    msg: str,
    timestamp: None | float = None,
    *,
    reference: None | tuple[float, float] = None,
) -> Message: ...


@overload
def decode(
    msg: list[str] | pd.Series,
    timestamp: None | Sequence[float] | pd.Series = None,
    *,
    reference: None | tuple[float, float] = None,
    batch: int = 1000,
) -> list[Message]: ...


def decode(
    msg: str | list[str] | pd.Series,
    timestamp: None | float | Sequence[float] | pd.Series = None,
    *,
    reference: None | tuple[float, float] = None,
    batch: int = 1000,
) -> Message | list[Message]:
    if isinstance(msg, str):
        if reference is not None:
            payload = decode_1090_with_reference(msg, reference)
        else:
            payload = decode_1090(msg)

    else:
        if timestamp is not None and isinstance(timestamp, (int, float)):
            raise ValueError(
                "`timestamp` parameter must be a sequence of float"
            )
        if timestamp is not None and len(timestamp) != len(msg):
            raise ValueError("`msg` and `timestamp` must be of the same length")

        batches = list(batched(msg, batch))
        if timestamp is None:
            if reference is not None:
                raise ValueError(
                    "Provide timestamps in order to fully decode positions"
                )
            payload = decode_1090_vec(batches)
        else:
            ts = list(batched(timestamp, batch))
            payload = decode_1090t_vec(batches, ts, reference)

    return pickle.loads(bytes(payload))  # type: ignore[no-any-return]


@overload
def flarm(
    msg: str,
    timestamp: int,
    reference_latitude: float,
    reference_longitude: float,
    *,
    batch: int = 1000,
) -> Flarm: ...


@overload
def flarm(
    msg: Sequence[str],
    timestamp: Sequence[int],
    reference_latitude: Sequence[float],
    reference_longitude: Sequence[float],
    *,
    batch: int = 1000,
) -> list[Flarm]: ...


def flarm(
    msg: str | Sequence[str],
    timestamp: int | Sequence[int],
    reference_latitude: float | Sequence[float],
    reference_longitude: float | Sequence[float],
    *,
    batch: int = 1000,
) -> Flarm | list[Flarm]:
    if isinstance(msg, str):
        assert isinstance(timestamp, (int, float))
        assert isinstance(reference_latitude, (int, float))
        assert isinstance(reference_longitude, (int, float))
        payload = decode_flarm(
            msg,
            timestamp,
            reference_latitude,
            reference_longitude,
        )
    else:
        batches = list(batched(msg, batch))
        assert not isinstance(timestamp, (int, float))
        assert not isinstance(reference_latitude, (int, float))
        assert not isinstance(reference_longitude, (int, float))

        t = list(batched(timestamp, batch))
        reflat = list(batched(reference_latitude, batch))
        reflon = list(batched(reference_longitude, batch))

        payload = decode_flarm_vec(batches, t, reflat, reflon)

    return pickle.loads(bytes(payload))  # type: ignore[no-any-return]



from pathlib import Path

import pandas as pd
import pytest

from rs1090 import flarm

root = Path(__file__)


def test_flarm() -> None:
    decoded = flarm(
        "7bf23810860b7eabb23952252fd4927024b21fd94e9e1ef416f0",
        1655274034,
        43.61924,
        5.11755,
    )
    assert decoded["icao24"] == "38f27b"
    assert decoded["is_icao24"]
    assert decoded["actype"] == "Glider"
    assert decoded["latitude"] == pytest.approx(43.61822)
    assert decoded["longitude"] == pytest.approx(5.117242)
    assert decoded["geoaltitude"] == 160
    assert decoded["vertical_speed"] == pytest.approx(-1.1)
    assert decoded["groundspeed"] == pytest.approx(0.7905694)
    assert decoded["track"] == pytest.approx(198.40446)
    assert not decoded["no_track"]
    assert not decoded["stealth"]
    assert decoded["gps"] == 3926


def test_full() -> None:
    data = pd.read_csv(
        root.parent.parent.parent / "crates/rs1090/data/flarm.csv",
    )

    decoded = flarm(
        data.rawmessage,
        data.timestamp.astype(int),
        data.sensorlatitude,
        data.sensorlongitude,
    )

    assert sum(1 for x in decoded if x["latitude"] > 47) == 0
    assert sum(1 for x in decoded if x["latitude"] < 40) == 0
    assert sum(1 for x in decoded if x["longitude"] > 12) == 0
    assert sum(1 for x in decoded if x["longitude"] < -2) == 0

    assert (
        sum(
            1
            for x in decoded
            if 43 < x["latitude"] < 44 and 4.5 < x["longitude"] < 5.5
        )
        / len(decoded)
        > 0.95
    )



# ruff: noqa: E402
# %%
import bench_pms  # type: ignore
import pandas as pd  # type: ignore
from pyModeS import c_common, py_common  # type: ignore

from rs1090 import decode

msg = "8DA05F219B06B6AF189400CBC33F"

decode(msg)
bench_pms.decode(msg, py_common)
bench_pms.decode(msg, c_common)

data = pd.read_csv(
    "../../crates/rs1090/data/long_flight.csv",
    names=["timestamp", "rawmsg"],
)

# %%
# %%timeit
# 1.69 s ± 60.7 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
decoded = decode(data.rawmsg.str[18:], data.timestamp, reference=(43.3, 1.35))

# %%
# %%timeit
# 1.46 s ± 30.3 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
decoded = decode(data.rawmsg.str[18:])

# %%
# %%timeit
# 4.15 s ± 92.8 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
decoded = decode(data.rawmsg.str[18:], batch=data.shape[0])

# %%
# %%timeit
# 9.27 s ± 154 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
decoded = [bench_pms.decode(msg, c_common) for msg in data.rawmsg.str[18:]]

# %%
# %%timeit
# 16 s ± 183 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
decoded = [bench_pms.decode(msg, py_common) for msg in data.rawmsg.str[18:]]


# %%
# cargo bench:  time:   [741.01 ms 755.14 ms 770.26 ms]

# %%
n = data.shape[0]
result = pd.DataFrame.from_records(
    [
        {"type": "rs1090 (Rust, default)", "time": 0.755, "std": 0.015},
        {"type": "rs1090 (Python, default)", "time": 1.46, "std": 0.03},
        {"type": "rs1090 (Python, single core)", "time": 4.15, "std": 0.093},
        {"type": "pyModeS (compiled)", "time": 9.27, "std": 0.154},
        {"type": "pyModeS (Python)", "time": 16, "std": 0.183},
    ]
)

# %%

import altair as alt  # type: ignore

chart = (
    alt.Chart(result)
    .mark_bar()
    .encode(
        alt.X(
            "time",
            axis=alt.Axis(format="~s"),
            title="Decoding time (in μs per message)",
        ),
        alt.Y("type", title=None, sort="-x"),
        alt.Color("type", legend=None),  # type: ignore
    )
    .transform_calculate(time=f"datum.time / {n}")
    .properties(width=500, height=100)
    .configure_axis(
        labelFont="Noto Sans",
        titleFont="Noto Sans",
        labelFontSize=13,
        titleFontSize=15,
        titleAnchor="end",
    )
)
chart.save("benchmark.svg")
chart

# %%



"""
Comprehensive corner case tests for rs1090 Python bindings.

These tests mirror the Rust tests in crates/rs1090/src/decode/bds/*.rs
and validate edge cases using real ADS-B messages from flight data.
"""

from pytest import approx

import rs1090


class TestBDS06SurfacePosition:
    """Tests for BDS 0,6 (Surface Position) movement field encoding.

    Validates the fix for movement codes 13-38 which should use 0.5 kt steps,
    not 0.25 kt steps as in the buggy version.
    """

    def test_movement_no_info(self) -> None:
        """Movement code 0 should return no groundspeed information."""
        msg = rs1090.decode("8c3944f8400002acb23cda192b95")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert "groundspeed" not in msg or msg["groundspeed"] is None

    def test_movement_stopped(self) -> None:
        """Movement code 1: aircraft stopped (0.0 kt)."""
        msg = rs1090.decode("903a33ff40100858d34ff3cce976")
        assert rs1090.is_df18(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == 0.0

    def test_movement_1_2kt_range(self) -> None:
        """Movement code 9: 1.0 kt (range 1.0-2.0 kt, 0.25 kt steps)."""
        msg = rs1090.decode("8c394c0f389b1667e947db7bb8bc")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == approx(1.0, abs=0.01)

    def test_movement_2_15kt_range_fix(self) -> None:
        """Movement code 25: 8.0 kt (validates 0.5 kt step fix).

        This message would decode to 5.0 kt with buggy 0.25 kt steps.
        Correct formula: 2.0 + (25-13)*0.5 = 8.0 kt
        """
        msg = rs1090.decode("8c3461cf399d6059814ea81483a9")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == approx(8.0, abs=0.01)

    def test_movement_2_15kt_range_fix_2(self) -> None:
        """Movement code 24: 7.5 kt (validates 0.5 kt step fix).

        This message would decode to 4.75 kt with buggy 0.25 kt steps.
        Correct formula: 2.0 + (24-13)*0.5 = 7.5 kt
        """
        msg = rs1090.decode("8c3461cf398d60597b4ea434c4d7")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == approx(7.5, abs=0.01)

    def test_movement_15_70kt_range(self) -> None:
        """Movement code 39: 15.0 kt (range 15-70 kt, 1.0 kt steps)."""
        msg = rs1090.decode("8c3461cf3a7f3059c94e5bf4e169")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == approx(15.0, abs=0.01)

    def test_movement_70_100kt_range(self) -> None:
        """Movement code 94: 70.0 kt (range 70-100 kt, 2.0 kt steps)."""
        msg = rs1090.decode("8c3950cf3dede47bac304d3b5122")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == approx(70.0, abs=0.01)

    def test_movement_100_175kt_range(self) -> None:
        """Movement code 109: 100.0 kt (range 100-175 kt, 5.0 kt steps)."""
        msg = rs1090.decode("8c3933203edde47b9e2ffa5e77b8")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == approx(100.0, abs=0.01)

    def test_movement_175kt_plus(self) -> None:
        """Movement code 124: 175.0 kt (≥175 kt)."""
        msg = rs1090.decode("8d3933203fcde2a84e39e1c6c5bc")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds06(msg)
        assert msg["groundspeed"] == approx(175.0, abs=0.01)

    def test_track_invalid(self) -> None:
        """Track status = 0 should return no track information."""
        msg = rs1090.decode("903a33ff40100858d34ff3cce976")
        assert rs1090.is_df18(msg)
        assert rs1090.is_bds06(msg)
        # Track should be None or not present when status is invalid
        assert "track" not in msg or msg["track"] is None


class TestBDS09AirborneVelocity:
    """Tests for BDS 0,9 (Airborne Velocity) vertical rate encoding.

    Validates sign bit handling and edge cases for vertical rate.
    """

    def test_vertical_rate_positive_64(self) -> None:
        """Vertical rate +64 ft/min (minimum positive rate)."""
        msg = rs1090.decode("8d3461cf9908388930080f948ea1")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds09(msg)
        assert msg["vertical_rate"] == 64
        assert msg["vrate_src"] == "GNSS"

    def test_vertical_rate_positive_128(self) -> None:
        """Vertical rate +128 ft/min."""
        msg = rs1090.decode("8d3461cf9908558e100c1071eb67")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds09(msg)
        assert msg["vertical_rate"] == 128
        assert msg["vrate_src"] == "GNSS"

    def test_vertical_rate_positive_960(self) -> None:
        """Vertical rate +960 ft/min."""
        msg = rs1090.decode("8d3461cf99085a8f10400f80e6ac")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds09(msg)
        assert msg["vertical_rate"] == 960
        assert msg["vrate_src"] == "GNSS"

    def test_vertical_rate_negative_64(self) -> None:
        """Vertical rate -64 ft/min (minimum negative rate)."""
        msg = rs1090.decode("8d394c0f990c4932780838866883")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds09(msg)
        assert msg["vertical_rate"] == -64
        assert msg["vrate_src"] == "GNSS"

    def test_vertical_rate_sign_bit(self) -> None:
        """Verify sign bit handling: same magnitude, opposite signs."""
        msg_pos = rs1090.decode("8d3461cf9908388930080f948ea1")  # +64
        msg_neg = rs1090.decode("8d394c0f990c4932780838866883")  # -64

        assert rs1090.is_df17(msg_pos)
        assert rs1090.is_df17(msg_neg)
        assert rs1090.is_bds09(msg_pos)
        assert rs1090.is_bds09(msg_neg)

        assert msg_pos["vertical_rate"] == 64
        assert msg_neg["vertical_rate"] == -64
        assert msg_pos["vertical_rate"] == -msg_neg["vertical_rate"]

    def test_vrate_source_gnss(self) -> None:
        """Vertical rate from GNSS source."""
        msg = rs1090.decode("8d3461cf9908388930080f948ea1")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds09(msg)
        assert msg["vrate_src"] == "GNSS"

    def test_vrate_source_barometric(self) -> None:
        """Vertical rate from barometric source."""
        msg = rs1090.decode("8D485020994409940838175B284F")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds09(msg)
        assert msg["vrate_src"] == "barometric"

    def test_geo_minus_baro_positive(self) -> None:
        """GNSS altitude above barometric (+550 ft)."""
        msg = rs1090.decode("8D485020994409940838175B284F")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds09(msg)
        assert msg["geo_minus_baro"] == 550


class TestBDS05AirbornePosition:
    """Tests for BDS 0,5 (Airborne Position) altitude encoding.

    Validates altitude decoding including negative values for airports
    below sea level (e.g., Amsterdam EHAM at -11 ft MSL).
    """

    def test_negative_altitude_325ft(self) -> None:
        """Altitude -325 ft (below sea level)."""
        msg = rs1090.decode("8d484fde5803b647ecec4fcdd74f")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds05(msg)
        assert msg["altitude"] == -325

    def test_negative_altitude_300ft(self) -> None:
        """Altitude -300 ft (below sea level)."""
        msg = rs1090.decode("8d4845575803c647bcec2a980abc")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds05(msg)
        assert msg["altitude"] == -300

    def test_negative_altitude_275ft(self) -> None:
        """Altitude -275 ft (below sea level)."""
        msg = rs1090.decode("8d3424d25803d64c18ee03351f89")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds05(msg)
        assert msg["altitude"] == -275

    def test_zero_altitude(self) -> None:
        """Altitude 0 ft (sea level)."""
        msg = rs1090.decode("8d4401e458058645a8ea90496290")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds05(msg)
        assert msg["altitude"] == 0

    def test_small_positive_altitude_25ft(self) -> None:
        """Altitude 25 ft (low altitude)."""
        msg = rs1090.decode("8d346355580596459cea86756acc")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds05(msg)
        assert msg["altitude"] == 25

    def test_positive_altitude_1000ft(self) -> None:
        """Altitude 1000 ft."""
        msg = rs1090.decode("8d346355580b064116e70a269f97")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds05(msg)
        assert msg["altitude"] == 1000

    def test_positive_altitude_5000ft(self) -> None:
        """Altitude 5000 ft."""
        msg = rs1090.decode("8d343386581f06318ad4fecab734")
        assert rs1090.is_df17(msg)
        assert rs1090.is_bds05(msg)
        assert msg["altitude"] == 5000

    # Note: test_altitude_all_zeros is skipped because we don't have a real
    # message with altitude field = 0x000 in our sample data.
    # The Rust tests use a synthetic message for this edge case.



from __future__ import annotations

from typing import Literal, TypedDict, Union, overload

from typing_extensions import (
    NotRequired,  # python <3.11
    TypeGuard,  # python <3.10
)


class BDS05(TypedDict):
    bds: Literal["05"]
    NUCp: int
    NICb: int
    altitude: int
    source: str
    parity: Literal["odd", "even"]
    lat_cpr: int
    lon_cpr: int
    latitude: NotRequired[float]
    longitude: NotRequired[float]


class BDS10(TypedDict):
    bds: Literal["10"]
    config: bool
    ovc: bool
    acas: bool
    subnet: int
    level5: bool
    mode_s: bool
    identification: bool
    squitter: bool
    sic: bool
    gicb: bool
    acas_hybrid: bool
    acas_ra: bool
    dte: int


class BDS17(TypedDict):
    bds: Literal["17"]
    bds05: NotRequired[Literal[True]]
    bds06: NotRequired[Literal[True]]
    bds07: NotRequired[Literal[True]]
    bds08: NotRequired[Literal[True]]
    bds09: NotRequired[Literal[True]]
    bds0a: NotRequired[Literal[True]]
    bds20: Literal[True]
    bds40: NotRequired[Literal[True]]
    bds41: NotRequired[Literal[True]]
    bds42: NotRequired[Literal[True]]
    bds43: NotRequired[Literal[True]]
    bds44: NotRequired[Literal[True]]
    bds45: NotRequired[Literal[True]]
    bds48: NotRequired[Literal[True]]
    bds50: NotRequired[Literal[True]]
    bds51: NotRequired[Literal[True]]
    bds52: NotRequired[Literal[True]]
    bds53: NotRequired[Literal[True]]
    bds54: NotRequired[Literal[True]]
    bds55: NotRequired[Literal[True]]
    bds56: NotRequired[Literal[True]]
    bds5f: NotRequired[Literal[True]]
    bds60: NotRequired[Literal[True]]


class BDS20(TypedDict):
    bds: Literal["20"]
    callsign: str
    icao24: str


class BDS21(TypedDict):
    bds: Literal["21"]
    registration: None | str
    airline: None | str


class BDS30(TypedDict):
    bds: Literal["30"]
    issued_ra: bool
    corrective: NotRequired[bool]
    downward_sense: NotRequired[bool]
    increased_rate: NotRequired[bool]
    sense_reversal: NotRequired[bool]
    altitude_crossing: NotRequired[bool]
    positive: NotRequired[bool]
    no_below: NotRequired[bool]
    no_above: NotRequired[bool]
    no_left: NotRequired[bool]
    no_right: NotRequired[bool]
    terminated: NotRequired[bool]
    multiple: NotRequired[bool]
    threat_identity: NotRequired[str]
    threat_altitude: NotRequired[int]
    threat_range: NotRequired[float]
    threat_bearing: NotRequired[int]


class BDS40(TypedDict):
    bds: Literal["40"]
    selected_mcp: NotRequired[int]
    selected_fms: NotRequired[int]
    barometric_setting: NotRequired[float]
    target_source: Literal[
        "AircraftAltitude", "FcpMcuSelectedAltitude", "FmsSelectedAltitude"
    ]


class BDS44(TypedDict):
    bds: Literal["44"]
    issued_ra: bool
    terminated: bool
    multiple: bool
    wind_speed: None | int
    wind_direction: None | float
    temperature: float
    pressure: None | int
    turbulence: Literal[None, "Nil", "Light", "Moderate", "Severe"]
    humidity: None | float


class BDS45(TypedDict):
    bds: Literal["45"]
    turbulence: Literal[None, "Nil", "Light", "Moderate", "Severe"]
    wind_shear: Literal[None, "Nil", "Light", "Moderate", "Severe"]
    icing: Literal[None, "Nil", "Light", "Moderate", "Severe"]
    wake_vortex: Literal[None, "Nil", "Light", "Moderate", "Severe"]
    static_temperature: float
    static_pressure: int
    radio_hein: None | int


class BDS50(TypedDict):
    bds: Literal["50"]
    roll: None | float
    track: None | float
    groundspeed: None | int
    track_rate: None | float
    TAS: None | int


class BDS60(TypedDict):
    bds: Literal["60"]
    heading: NotRequired[float]
    IAS: NotRequired[int]
    Mach: NotRequired[float]
    vrate_barometric: NotRequired[int]
    vrate_inertial: NotRequired[int]


class DF0(TypedDict):
    timestamp: float
    df: Literal["0"]
    altitude: int
    icao24: str


class DF4(TypedDict):
    timestamp: float
    df: Literal["4"]
    altitude: int
    icao24: str


class DF5(TypedDict):
    timestamp: float
    df: Literal["5"]
    squawk: str
    icao24: str


class DF11(TypedDict):
    timestamp: float
    df: Literal["11"]
    capability: str
    icao24: str


class DF16(TypedDict):
    timestamp: float
    df: Literal["16"]
    vs: int
    sl: int
    ri: int
    altitude: int
    icao24: str


class DF17_BDS05(TypedDict):
    timestamp: float
    df: Literal["17"]
    icao24: str
    bds: Literal["05"]
    NUCp: int
    NICb: int
    altitude: int
    source: str
    parity: Literal["odd", "even"]
    lat_cpr: int
    lon_cpr: int
    latitude: NotRequired[float]
    longitude: NotRequired[float]


class DF17_BDS06(TypedDict):
    timestamp: float
    df: Literal["17"]
    icao24: str
    bds: Literal["06"]
    NUCp: int
    groundspeed: None | float
    track: None | float
    parity: Literal["odd", "even"]
    lat_cpr: int
    lon_cpr: int
    latitude: NotRequired[float]
    longitude: NotRequired[float]


class DF17_BDS08(TypedDict):
    timestamp: float
    df: Literal["17"]
    icao24: str
    bds: Literal["08"]
    wake_vortex: Literal[
        "n/a",
        "Surface emergency vehicle",
        "Surface service vehicle",
        "Obstruction",
        "Glider",
        "Lighter than air",
        "Parachutist",
        "Ultralight",
        "UAM",
        "Space",
        "<7000kg",
        "34,000kg",
        "<136,000kg",
        "High vortex",
        "Heavy",
        "High performance",
        "Rotorcraft",
    ]
    callsign: str


class DF17_BDS09(TypedDict):
    timestamp: float
    df: Literal["17"]
    icao24: str
    bds: Literal["09"]
    NACv: int
    groundspeed: NotRequired[float]
    TAS: NotRequired[float]
    IAS: NotRequired[float]
    track: NotRequired[float]
    heading: NotRequired[float]
    vrate_src: str
    vertical_rate: int
    geo_minus_baro: None | int


class DF17_BDS61(TypedDict):
    timestamp: float
    df: str
    icao24: str
    bds: Literal["61"]
    subtype: Literal["emergency_priority", "acas_ra"]
    emergency_state: Literal[
        "none",
        "general",
        "medical",
        "minimum_fuel",
        "no_communication",
        "unlawful_interference",
        "downed_aircraft",
        "reserved",
    ]
    squawk: str


class DF17_BDS62(TypedDict):
    timestamp: float
    df: Literal["17"]
    icao24: str
    bds: Literal["62"]
    source: str
    selected_altitude: int
    barometric_setting: float
    selected_heading: NotRequired[float]
    NACp: int
    autopilot: NotRequired[bool]
    vnav_mode: NotRequired[bool]
    alt_hold: NotRequired[bool]
    approach_mode: NotRequired[bool]
    tcas_operational: bool
    lnav_mode: NotRequired[bool]


class DF17_BDS65(TypedDict):
    timestamp: float
    df: Literal["17"]
    icao24: str
    bds: Literal["65"]
    version: Literal["1", "2"]
    # NIC supplement A
    NICa: int
    # NIC Supplement bit (NICs)
    NICs: NotRequired[int]
    # Navigation Accuracy Category for position (NACp)
    NACp: int
    #  Geometry Vertical Accuracy (GVA)
    GVA: NotRequired[int]
    # Barometric Altitude Integrity (BAI)
    BAI: int
    # Barometric Altitude Quality (BAQ)
    BAQ: NotRequired[int]
    # Track Angle or Heading
    TAH: NotRequired[int]
    # 1 for magnetic, 0 for true north
    HRD: Literal[0, 1]
    # Surveillance Integrity Level (SIL)
    SIL: int
    SILs: int


class DF17_Unknown(TypedDict):
    timestamp: float
    df: Literal["17"]
    icao24: str
    bds: Literal["?"]


class DF18_BDS06(TypedDict):
    timestamp: float
    df: Literal["18"]
    tisb: Literal[
        "ADSB_ES_NT",
        "ADSB_ES_NT_ALT",
        "TISB_FINE",
        "TISB_COARSE",
        "TISB_MANAGE",
        "TISB_ADSB_RELAY",
        "TISB_ADSB",
        "Reserved",
    ]
    icao24: str
    bds: Literal["06"]
    NUCp: int
    groundspeed: None | float
    track: None | float
    parity: Literal["odd", "even"]
    lat_cpr: int
    lon_cpr: int
    latitude: NotRequired[float]
    longitude: NotRequired[float]


class DF18_BDS08(TypedDict):
    timestamp: float
    df: Literal["18"]
    tisb: Literal[
        "ADSB_ES_NT",
        "ADSB_ES_NT_ALT",
        "TISB_FINE",
        "TISB_COARSE",
        "TISB_MANAGE",
        "TISB_ADSB_RELAY",
        "TISB_ADSB",
        "Reserved",
    ]
    icao24: str
    bds: Literal["08"]
    wake_vortex: Literal[
        "n/a",
        "Surface emergency vehicle",
        "Surface service vehicle",
        "Obstruction",
        "Glider",
        "Lighter than air",
        "Parachutist",
        "Ultralight",
        "UAM",
        "Space",
        "<7000kg",
        "34,000kg",
        "<136,000kg",
        "High vortex",
        "Heavy",
        "High performance",
        "Rotorcraft",
    ]
    callsign: str


class DF18_BDS65(TypedDict):
    timestamp: float
    df: Literal["18"]
    tisb: Literal[
        "ADSB_ES_NT",
        "ADSB_ES_NT_ALT",
        "TISB_FINE",
        "TISB_COARSE",
        "TISB_MANAGE",
        "TISB_ADSB_RELAY",
        "TISB_ADSB",
        "Reserved",
    ]
    icao24: str
    bds: Literal["65"]
    version: Literal["1", "2"]
    # NIC supplement A
    NICa: int
    # NIC Supplement bit (NICs)
    NICs: NotRequired[int]
    # Navigation Accuracy Category for position (NACp)
    NACp: int
    #  Geometry Vertical Accuracy (GVA)
    GVA: NotRequired[int]
    # Barometric Altitude Integrity (BAI)
    BAI: int
    # Barometric Altitude Quality (BAQ)
    BAQ: NotRequired[int]
    # Track Angle or Heading
    TAH: NotRequired[int]
    # 1 for magnetic, 0 for true north
    HRD: Literal[0, 1]
    # Surveillance Integrity Level (SIL)
    SIL: int
    SILs: int


class DF18_Unknown(TypedDict):
    timestamp: float
    df: Literal["18"]
    tisb: Literal[
        "ADSB_ES_NT",
        "ADSB_ES_NT_ALT",
        "TISB_FINE",
        "TISB_COARSE",
        "TISB_MANAGE",
        "TISB_ADSB_RELAY",
        "TISB_ADSB",
        "Reserved",
    ]
    icao24: str
    bds: Literal["?"]


class DF20(TypedDict):
    timestamp: float
    df: Literal["20"]
    altitude: int
    icao24: str
    bds05: None | BDS05
    bds10: None | BDS10
    bds17: None | BDS17
    # bds18: None | BDS18
    # bds19: None | BDS19
    bds20: None | BDS20
    bds21: None | BDS21
    bds30: None | BDS30
    bds40: None | BDS40
    bds44: None | BDS44
    bds45: None | BDS45
    bds50: None | BDS50
    bds60: None | BDS60


class DF21(TypedDict):
    timestamp: float
    df: Literal["21"]
    squawk: str
    icao24: str
    bds10: None | BDS10
    bds17: None | BDS17
    # bds18: None | BDS18
    # bds19: None | BDS19
    bds20: None | BDS20
    bds21: None | BDS21
    bds30: None | BDS30
    bds40: None | BDS40
    bds44: None | BDS44
    bds45: None | BDS45
    bds50: None | BDS50
    bds60: None | BDS60


DF17 = Union[
    DF17_Unknown,
    DF17_BDS05,
    DF17_BDS06,
    DF17_BDS08,
    DF17_BDS09,
    DF17_BDS61,
    DF17_BDS62,
    DF17_BDS65,
]

DF18 = Union[
    DF18_Unknown,
    DF18_BDS06,
    DF18_BDS08,
    DF18_BDS65,
]


Message = Union[DF0, DF4, DF5, DF11, DF16, DF17, DF18, DF20, DF21]


def is_df0(message: Message) -> TypeGuard[DF0]:
    return message["df"] == "0"


def is_df4(message: Message) -> TypeGuard[DF4]:
    return message["df"] == "4"


def is_df5(message: Message) -> TypeGuard[DF5]:
    return message["df"] == "5"


def is_df11(message: Message) -> TypeGuard[DF11]:
    return message["df"] == "11"


def is_df16(message: Message) -> TypeGuard[DF16]:
    return message["df"] == "16"


def is_df17(message: Message) -> TypeGuard[DF17]:
    return message["df"] == "17"


def is_df18(message: Message) -> TypeGuard[DF18]:
    return message["df"] == "18"


def is_df20(message: Message) -> TypeGuard[DF20]:
    return message["df"] == "20"


def is_df21(message: Message) -> TypeGuard[DF21]:
    return message["df"] == "21"


def is_bds05(message: DF17) -> TypeGuard[DF17_BDS05]:
    return message.get("bds", None) == "05" or "bds05" in message


@overload
def is_bds06(message: DF17) -> TypeGuard[DF17_BDS06]: ...
@overload
def is_bds06(message: DF18) -> TypeGuard[DF18_BDS06]: ...
def is_bds06(message: DF17 | DF18) -> TypeGuard[DF17_BDS06 | DF18_BDS06]:
    return message.get("bds", None) == "06" or "bds06" in message


@overload
def is_bds08(message: DF17) -> TypeGuard[DF17_BDS08]: ...
@overload
def is_bds08(message: DF18) -> TypeGuard[DF18_BDS08]: ...
def is_bds08(message: DF17 | DF18) -> TypeGuard[DF17_BDS08 | DF18_BDS08]:
    return message.get("bds", None) == "08" or "bds08" in message


def is_bds09(message: DF17) -> TypeGuard[DF17_BDS09]:
    return message.get("bds", None) == "09" or "bds09" in message


def is_bds10(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "10" or "bds10" in message


def is_bds17(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "17" or "bds17" in message


def is_bds20(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "20" or "bds20" in message


def is_bds30(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "30" or "bds30" in message


def is_bds40(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "40" or "bds40" in message


def is_bds44(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "44" or "bds44" in message


def is_bds50(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "50" or "bds50" in message


def is_bds60(message: DF20 | DF21) -> bool:
    return message.get("bds", None) == "60" or "bds60" in message


def is_bds61(message: DF17) -> TypeGuard[DF17_BDS61]:
    return message.get("bds", None) == "61" or "bds61" in message


def is_bds62(message: DF17) -> TypeGuard[DF17_BDS62]:
    return message.get("bds", None) == "62" or "bds62" in message


def is_bds65(message: DF17) -> TypeGuard[DF17_BDS65]:
    return message.get("bds", None) == "65" or "bds65" in message


class Flarm(TypedDict):
    timestamp: int
    reference_lat: float
    reference_lon: float
    icao24: str
    is_icao24: bool
    actype: Literal[
        "Unknown",
        "Glider",
        "Towplane",
        "Helicopter",
        "Parachute",
        "DropPlane",
        "Hangglider",
        "Paraglider",
        "Aircraft",
        "Jet",
        "UFO",
        "Balloon",
        "Airship",
        "UAV",
        "Reserved",
        "StaticObstacle",
    ]
    latitude: float
    longitude: float
    geoaltitude: int
    vertical_speed: float
    groundspeed: float
    track: float
    no_track: bool
    stealth: bool
    gps: int



from pathlib import Path

import pandas as pd

from rs1090 import decode

root = Path(__file__)


def test_full() -> None:
    data = pd.read_csv(
        root.parent.parent.parent / "crates/rs1090/data/long_flight.csv",
        names=["timestamp", "rawmsg"],
    )

    decoded = decode(
        data.rawmsg.str[18:],
        data.timestamp,
        reference=(43.3, 1.35),
    )

    assert data.shape[0] == len(decoded)



from pytest import approx

import rs1090


def test_icao24() -> None:
    assert rs1090.decode("8D406B902015A678D4D220AA4BDA")["icao24"] == "406b90"


def test_wake_vortex() -> None:
    msg = rs1090.decode("8D406B902015A678D4D220AA4BDA")
    assert rs1090.is_df17(msg)
    assert rs1090.is_bds08(msg)
    assert msg["wake_vortex"] == "n/a"


def test_adsb_callsign() -> None:
    msg = rs1090.decode("8D406B902015A678D4D220AA4BDA")
    assert rs1090.is_df17(msg)
    assert rs1090.is_bds08(msg)
    assert msg["callsign"] == "EZY85MH"


def test_adsb_alt() -> None:
    msg = rs1090.decode("8D40058B58C901375147EFD09357")
    assert rs1090.is_df17(msg)
    assert rs1090.is_bds05(msg)
    assert msg["altitude"] == 39000


def test_adsb_velocity() -> None:
    msg = rs1090.decode("8D485020994409940838175B284F")
    assert rs1090.is_df17(msg)
    assert rs1090.is_bds09(msg)
    assert msg["groundspeed"] == approx(159.2, rel=1e-3)
    assert msg["vertical_rate"] == -832
    assert msg["track"] == approx(182.88, rel=1e-3)
    assert msg["geo_minus_baro"] == 550

    msg = rs1090.decode("8DA05F219B06B6AF189400CBC33F")
    assert rs1090.is_df17(msg)
    assert rs1090.is_bds09(msg)
    assert msg["TAS"] == 375
    assert msg["vertical_rate"] == -2304
    assert msg["heading"] == approx(243.98, rel=1e-3)


def test_adsb_emergency() -> None:
    msg = rs1090.decode("8DA2C1B6E112B600000000760759")
    assert rs1090.is_df17(msg)
    assert rs1090.is_bds61(msg)
    assert msg["emergency_state"] == "none"
    assert msg["squawk"] == "6513"


def test_adsb_target_state_status() -> None:
    msg = rs1090.decode("8DA05629EA21485CBF3F8CADAEEB")
    assert rs1090.is_df17(msg)
    assert rs1090.is_bds62(msg)
    assert msg["selected_altitude"] == 17000
    assert msg["source"] == "MCP/FCU"
    assert msg["barometric_setting"] == approx(1012.8)
    assert msg["selected_heading"] == approx(66.8, 0.1)
    assert msg["autopilot"]
    assert msg["vnav_mode"]
    assert not msg["alt_hold"]
    assert not msg["approach_mode"]
    assert msg["lnav_mode"]
    assert msg["tcas_operational"]


def test_adsb_with_position() -> None:
    msg = rs1090.decode(
        "90343652300003eeda6de84f1ad2", reference=(40.48, -3.56)
    )
    assert rs1090.is_df18(msg)
    assert rs1090.is_bds06(msg)
    assert msg["latitude"] == approx(40.4749)
    assert msg["longitude"] == approx(-3.57068)



from pytest import approx

import rs1090


def test_bds20() -> None:
    msg = rs1090.decode("A000083E202CC371C31DE0AA1CCF")
    assert rs1090.is_df20(msg)
    bds20 = msg["bds20"]
    assert bds20 is not None
    assert bds20["callsign"] == "KLM1017"
    msg = rs1090.decode("A0001993202422F2E37CE038738E")
    assert rs1090.is_df20(msg)
    bds20 = msg["bds20"]
    assert bds20 is not None
    assert bds20["callsign"] == "IBK2873"


def test_bds40() -> None:
    msg = rs1090.decode("A000029C85E42F313000007047D3")
    assert rs1090.is_df20(msg)
    bds40 = msg["bds40"]
    assert bds40 is not None
    assert bds40["selected_mcp"] == 3000
    assert bds40["selected_fms"] == 3000
    assert bds40["barometric_setting"] == 1020


def test_bds50() -> None:
    msg = rs1090.decode("A000139381951536E024D4CCF6B5")
    assert rs1090.is_df20(msg)
    bds50 = msg["bds50"]
    assert bds50 is not None
    assert bds50["roll"] == approx(2.11, rel=1e-3)
    assert bds50["track"] == approx(114.2578)
    assert bds50["groundspeed"] == 438
    assert bds50["track_rate"] == 0.125
    assert bds50["TAS"] == 424


def test_bds60() -> None:
    msg = rs1090.decode("A00004128F39F91A7E27C46ADC21")
    assert rs1090.is_df20(msg)
    bds60 = msg["bds60"]
    assert bds60 is not None
    assert bds60["heading"] == approx(42.71484)
    assert bds60["IAS"] == 252
    assert bds60["Mach"] == 0.42
    assert bds60["vrate_barometric"] == -1920
    assert bds60["vrate_inertial"] == -1920



from rs1090 import aircraft_information


def test_info() -> None:
    info = aircraft_information("39b415")
    assert info == {
        "country": "France",
        "registration": "F-HNAV",
        "pattern": "^F-",
        "icao24": "39b415",
        "flag": "🇫🇷",
    }
